@@ -25,7 +25,7 @@ and binding sites.
 .. autosummary::
    :toctree: generated/
 
-   api
+   LipidInteraction
 
 """
 
