@@ -16,6 +16,7 @@
 
 from collections import defaultdict
 from itertools import product
+from functools import partial
 import pickle
 import os
 import re
@@ -24,20 +25,18 @@ import mdtraj as md
 import numpy as np
 np.seterr(all='ignore')
 from scipy.sparse import coo_matrix
-from sklearn.decomposition import PCA
 import pandas as pd
 from tqdm import trange, tqdm
+from p_tqdm import p_map
 from ..func import cal_contact_residues
-from ..func import Duration, cal_koff
+from ..func import Duration
 from ..func import cal_lipidcount, cal_occupancy
 from ..func import get_node_list
-from ..func import collect_bound_poses, vectorize_poses, calculate_scores, write_bound_poses
-from ..func import cluster_DBSCAN, cluster_KMeans
-from ..func import calculate_site_surface_area
-from ..plot import plot_koff
+from ..func import collect_bound_poses
+from ..func import analyze_pose_wrapper, calculate_koff_wrapper, calculate_surface_area_wrapper
 from ..plot import plot_surface_area, plot_binding_site_data
 from ..plot import plot_residue_data, plot_corrcoef, plot_residue_data_logos
-from ..util import check_dir, write_PDB, write_pymol_script, sparse_corrcoef, rmsd, get_traj_info
+from ..util import check_dir, write_PDB, write_pymol_script, sparse_corrcoef, get_traj_info
 
 
 class LipidInteraction:
@@ -305,9 +304,9 @@ class LipidInteraction:
                 self._protein_residue_id = traj_info["protein_residue_id"]
                 self._residue_list = traj_info["residue_list"]
                 self._nresi_per_protein = len(self._residue_list)
-                self._duration = {residue_id: [] for residue_id in self._protein_residue_id}
-                self._occupancy = {residue_id: [] for residue_id in self._protein_residue_id}
-                self._lipid_count = {residue_id: [] for residue_id in self._protein_residue_id}
+                self._duration = dict()
+                self._occupancy = dict()
+                self._lipid_count = dict()
                 self._contact_residues_high = {residue_id: [] for residue_id in self._protein_residue_id}
                 self._contact_residues_low = {residue_id: [] for residue_id in self._protein_residue_id}
                 self._koff = np.zeros(self._nresi_per_protein)
@@ -323,6 +322,7 @@ class LipidInteraction:
                         traj_idx, len(traj_info["protein_residue_id"]), traj_idx - 1, len(self._protein_residue_id))
             ncol_per_protein = len(traj_info["lipid_residue_atomid_list"]) * traj.n_frames
             for protein_idx in np.arange(self._nprot, dtype=int):
+
                 for residue_id, residue_atom_indices in enumerate(
                         traj_info["protein_residue_atomid_list"][protein_idx]):
                     # calculate interaction per residue
@@ -338,6 +338,7 @@ class LipidInteraction:
                                 frame_id for frame_id, lipid_id in zip(frame_id_set_low, lipid_id_set_low)])
                     row.append([residue_id for dummy in np.arange(len(frame_id_set_low), dtype=int)])
                     data.append(dist_matrix[lipid_id_set_low, frame_id_set_low])
+
             ncol_start += ncol_per_protein * self._nprot
         # calculate correlation coefficient matrix
         row = np.concatenate(row)
@@ -366,14 +367,13 @@ class LipidInteraction:
             selected_residue_id = self._protein_residue_id
         else:
             selected_residue_id = np.atleast_1d(residue_id)
-
-        for residue_id in selected_residue_id:
+        for residue_id in tqdm(selected_residue_id, desc="CALCULATE DURATION PER RESIDUE"):
             self._duration[residue_id] = [
                         Duration(self._contact_residues_low[residue_id][(traj_idx*self._nprot)+protein_idx],
                                  self._contact_residues_high[residue_id][(traj_idx*self._nprot)+protein_idx],
                                  self._timesteps[traj_idx]).cal_durations()
                         for traj_idx in np.arange(len(self.trajfile_list))
-                        for protein_idx in np.arange(self._nprot, dtype=int)]
+                        for protein_idx in np.arange(self._nprot, dtype=int)]       
         self.dataset["Duration"] = [np.mean(np.concatenate(self._duration[residue_id]))
                                     if len(self._duration[residue_id]) > 0 else 0
                                     for residue_id in self._protein_residue_id]
@@ -403,7 +403,7 @@ class LipidInteraction:
             selected_residue_id = self._protein_residue_id
         else:
             selected_residue_id = np.atleast_1d(residue_id)
-        for residue_id in selected_residue_id:
+        for residue_id in tqdm(selected_residue_id, desc="CALCULATE OCCUPANCY"):
             self._occupancy[residue_id] = [cal_occupancy(self._contact_residues_low[residue_id][(traj_idx*self._nprot)+protein_idx])
                                           for traj_idx in np.arange(len(self.trajfile_list))
                                           for protein_idx in np.arange(self._nprot, dtype=int)]
@@ -439,23 +439,23 @@ class LipidInteraction:
             selected_residue_id = self._protein_residue_id
         else:
             selected_residue_id = np.atleast_1d(residue_id)
-        for residue_id in selected_residue_id:
+        for residue_id in tqdm(selected_residue_id, desc="CALCULATE RESIDUE LIPIDCOUNT"):
             self._lipid_count[residue_id] = [cal_lipidcount(self._contact_residues_low[residue_id][(traj_idx*self._nprot)+protein_idx])
                                            for traj_idx in np.arange(len(self.trajfile_list))
                                            for protein_idx in np.arange(self._nprot, dtype=int)]
-            self.dataset["Lipid Count"] = [np.mean(self._lipid_count[residue_id])
+        self.dataset["Lipid Count"] = [np.mean(self._lipid_count[residue_id])
+                                       if len(self._lipid_count[residue_id]) > 0 else 0
+                                       for residue_id in self._protein_residue_id]
+        self.dataset["Lipid Count std"] = [np.std(self._lipid_count[residue_id])
                                            if len(self._lipid_count[residue_id]) > 0 else 0
                                            for residue_id in self._protein_residue_id]
-            self.dataset["Lipid Count std"] = [np.std(self._lipid_count[residue_id])
-                                               if len(self._lipid_count[residue_id]) > 0 else 0
-                                               for residue_id in self._protein_residue_id]
         if len(selected_residue_id) == 1:
             self._lipid_count[residue_id]
         else:
             return [self._lipid_count[residue_id] for residue_id in selected_residue_id]
 
     def compute_residue_koff(self, residue_id=None, nbootstrap=10, initial_guess=[1., 1., 1., 1.],
-                             save_dir=None, plot_data=True, fig_close=True):
+                             save_dir=None, plot_data=True, fig_close=True, num_cpus=None):
         """Calculate interaction koff and residence time for residues.
 
         Parameters
@@ -467,6 +467,7 @@ class LipidInteraction:
         print_data : bool, default=True
         plot_data : bool, default=True
         fig_close : bool, default=True
+        num_cpus : int or None, default=None
 
         Returns
         ---------
@@ -502,31 +503,25 @@ class LipidInteraction:
         if not same_timestep:
             warnings.warn(
                 "Trajectories have different timesteps. This will impair the accuracy of koff calculation!")
+        if plot_data:
+            fn_set = [os.path.join(koff_dir, "{}.pdf".format(residue_name_set[residue_id]))
+                      for residue_id in selected_residue_id]
+        else:
+            fn_set = [False for dummy in selected_residue_id]
+                
+        returned_values = p_map(partial(calculate_koff_wrapper, t_total=t_total, timestep=timestep, nbootstrap=nbootstrap,
+                                        initial_guess=initial_guess, plot_data=plot_data, timeunit=self._timeunit,
+                                        fig_close=fig_close),
+                                [np.concatenate(self._duration[residue_id]) for residue_id in selected_residue_id],
+                                [residue_name_set[residue_id] for residue_id in selected_residue_id],
+                                fn_set, num_cpus=num_cpus, desc="CALCULATE KOFF FOR RESIDUES")
 
-        for residue_id in tqdm(selected_residue_id,desc="CALCULATE KOFF FOR RESIDUES", total=len(selected_residue_id)):
-            durations = np.concatenate(self._duration[residue_id])
-            residue = residue_name_set[residue_id]
-            if np.sum(durations) == 0:
-                self._koff[residue_id] = 0
-                self._res_time[residue_id] = 0
-                self._r_squared[residue_id] = 0
-                self._koff_boot[residue_id] = 0
-                self._r_squared_boot[residue_id] = 0
-            else:
-                with warnings.catch_warnings():
-                    warnings.simplefilter("ignore")
-                    koff, restime, properties = cal_koff(durations, t_total, timestep, nbootstrap, initial_guess)
-                self._koff[residue_id] = koff
-                self._res_time[residue_id] = restime
-                self._r_squared[residue_id] = properties["r_squared"]
-                self._koff_boot[residue_id] = np.mean(properties["ks_boot_set"], axis=0)[0]
-                self._r_squared_boot[residue_id] = np.mean(properties["r_squared_boot_set"])
-                if plot_data:
-                    text = self._format_koff_text(properties)
-                    plot_koff(durations, properties["delta_t_list"], properties["survival_rates"],
-                              properties["n_fitted"], survival_rates_bootstraps=properties["survival_rates_boot_set"],
-                              fig_fn=os.path.join(koff_dir, "{}.pdf".format(residue)), title=residue,
-                              timeunit=self._timeunit, t_total=t_total, text=text, fig_close=fig_close)
+        for residue_id, returned_value in zip(selected_residue_id, returned_values):
+            self._koff[residue_id] = returned_value[0]
+            self._res_time[residue_id] = returned_value[1]
+            self._r_squared[residue_id] = returned_value[2]
+            self._koff_boot[residue_id] = returned_value[3]
+            self._r_squared_boot[residue_id] = returned_value[4]
         # update dataset
         self.dataset["Koff"] = self._koff
         self.dataset["Residence Time"] = self._res_time
@@ -572,9 +567,9 @@ class LipidInteraction:
             # update dataset
             self.dataset["Binding Site ID"] = residue_BS_identifiers
             # initialise variable for binding site interactions
-            self._duration_BS = {bs_id:[] for bs_id in np.arange(len(self._node_list))}
-            self._occupancy_BS = {bs_id:[] for bs_id in np.arange(len(self._node_list))}
-            self._lipid_count_BS = {bs_id:[] for bs_id in np.arange(len(self._node_list))}
+            self._duration_BS = dict()
+            self._occupancy_BS = dict()
+            self._lipid_count_BS = dict()
             self._koff_BS = np.zeros(len(self._node_list))
             self._koff_BS_boot = np.zeros(len(self._node_list))
             self._res_time_BS = np.zeros(len(self._node_list))
@@ -607,7 +602,7 @@ class LipidInteraction:
         self._check_calculation("Binding Site ID", self.compute_binding_nodes, print_data=False)
         selected_bs_id = np.atleast_1d(binding_site_id) if binding_site_id is not None \
             else np.arange(len(self._node_list), dtype=int)
-        for bs_id in selected_bs_id:
+        for bs_id in tqdm(selected_bs_id, desc="CALCULATE DURATION PER BINDING SITE"):
             nodes = self._node_list[bs_id]
             durations_BS = []
             for traj_idx in np.arange(len(self._trajfile_list), dtype=int):
@@ -651,7 +646,7 @@ class LipidInteraction:
         self._check_calculation("Binding Site ID", self.compute_binding_nodes, print_data=False)
         selected_bs_id = np.atleast_1d(binding_site_id) if binding_site_id is not None \
             else np.arange(len(self._node_list), dtype=int)
-        for bs_id in selected_bs_id:
+        for bs_id in tqdm(selected_bs_id, desc="CALCULATE OCCUPANCY PER BINDING SITE"):
             nodes = self._node_list[bs_id]
             occupancy_BS = []
             for traj_idx in np.arange(len(self._trajfile_list), dtype=int):
@@ -694,7 +689,7 @@ class LipidInteraction:
         self._check_calculation("Binding Site ID", self.compute_binding_nodes, print_data=False)
         selected_bs_id = np.atleast_1d(binding_site_id) if binding_site_id is not None \
             else np.arange(len(self._node_list), dtype=int)
-        for bs_id in selected_bs_id:
+        for bs_id in tqdm(selected_bs_id, desc="CALCULATE LIPIDCOUNT PER BINDING SITE"):
             nodes = self._node_list[bs_id]
             lipidcount_BS = []
             for traj_idx in np.arange(len(self._trajfile_list), dtype=int):
@@ -719,7 +714,7 @@ class LipidInteraction:
             return [self._lipid_count_BS[bs_id] for bs_id in selected_bs_id]
 
     def compute_site_koff(self, binding_site_id=None, nbootstrap=10, initial_guess=[1., 1., 1., 1.],
-                          save_dir=None, plot_data=True, fig_close=True):
+                          save_dir=None, plot_data=True, fig_close=True, num_cpus=None):
         """Calculate interactions koff and residence time for binding sites.
 
         Parameters
@@ -730,6 +725,7 @@ class LipidInteraction:
         save_dir : str, default=None
         plot_data : bool, default=True
         fig_close : bool, default=True
+        num_cpus : int or None, default=None
 
         Returns
         ---------
@@ -739,6 +735,7 @@ class LipidInteraction:
         """
         self._check_calculation("Residue", self.collect_residue_contacts)
         self._check_calculation("Binding Site ID", self.compute_binding_nodes, print_data=False)
+        self._check_calculation("Binding Site Duration", self.compute_site_duration)
         if plot_data:
             BS_dir = check_dir(save_dir, "Binding_Sites_koffs_{}".format(self._lipid)) if save_dir is not None \
                 else check_dir(self._save_dir, "Binding_Sites_koffs_{}".format(self._lipid))
@@ -759,24 +756,22 @@ class LipidInteraction:
         if not same_timestep:
             warnings.warn(
                 "Trajectories have different timesteps. This will impair the accuracy of koff calculation!")
-        # calculate koff for binding sites
-        for bs_id in tqdm(selected_bs_id, desc="CALCULATE KOFF FOR BINDING SITES", total=len(selected_bs_id)):
-            durations = np.concatenate(self._duration_BS[bs_id])
-            with warnings.catch_warnings():
-                warnings.simplefilter("ignore")
-                koff_BS, restime_BS, properties_BS = cal_koff(durations, t_total, timestep, nbootstrap, initial_guess)
-            self._koff_BS[bs_id] = koff_BS
-            self._koff_BS_boot[bs_id] = np.mean(properties_BS["ks_boot_set"], axis=0)[0]
-            self._res_time_BS[bs_id] = restime_BS
-            self._r_squared_BS[bs_id] = properties_BS["r_squared"]
-            self._r_squared_BS_boot[bs_id] = np.mean(properties_BS["r_squared_boot_set"])
-            if plot_data:
-                # plot BS koff and BS residence time
-                text = self._format_koff_text(properties_BS)
-                plot_koff(durations, properties_BS["delta_t_list"], properties_BS["survival_rates"],
-                          properties_BS["n_fitted"], survival_rates_bootstraps=properties_BS["survival_rates_boot_set"],
-                          fig_fn=os.path.join(BS_dir, f"BS_id{bs_id}.pdf"), title=f"Binding Site {bs_id}",
-                          timeunit=self._timeunit, t_total=t_total, text=text, fig_close=fig_close)
+        if plot_data:
+            fn_set = [os.path.join(BS_dir, f"BS_id{bs_id}.pdf").format(bs_id) for bs_id in selected_bs_id]
+        else:
+            fn_set = [False for dummy in selected_bs_id]
+        returned_values = p_map(partial(calculate_koff_wrapper, t_total=t_total, timestep=timestep, nbootstrap=nbootstrap,
+                                        initial_guess=initial_guess, plot_data=plot_data, timeunit=self._timeunit,
+                                        fig_close=fig_close),
+                                [np.concatenate(self._duration_BS[bs_id]) for bs_id in selected_bs_id],
+                                [f"Binding Site {bs_id}" for bs_id in selected_bs_id],
+                                fn_set, num_cpus=num_cpus, desc="CALCULATE KOFF FOR BINDING SITES")
+        for bs_id, returned_value in zip(selected_bs_id, returned_values):
+            self._koff_BS[bs_id] = returned_value[0]
+            self._res_time_BS[bs_id] = returned_value[1]
+            self._r_squared_BS[bs_id] = returned_value[2]
+            self._koff_BS_boot[bs_id] = returned_value[3]
+            self._r_squared_BS_boot[bs_id] = returned_value[4]
         # update dataset
         for data, column_name in zip(
                 [self._koff_BS, self._koff_BS_boot, self._res_time_BS, self._r_squared_BS, self._r_squared_BS_boot],
@@ -786,7 +781,6 @@ class LipidInteraction:
             for bs_id, nodes in enumerate(self._node_list):
                 data_per_residue[nodes] = data[bs_id]
             self.dataset[column_name] = data_per_residue
-
         if len(selected_bs_id) == 1:
             return self._koff_BS[bs_id], self._res_time_BS[bs_id]
         else:
@@ -794,8 +788,9 @@ class LipidInteraction:
                    [self._res_time_BS[bs_id] for bs_id in selected_bs_id]
 
     def analyze_bound_poses(self, binding_site_id=None, n_top_poses=3, pose_format="gro", score_weights=None,
-                            kde_bw=0.15, pca_component=0.95, plot_rmsd=True, save_dir=None,
-                            n_clusters="auto", eps=None, min_samples=None, metric="euclidean", fig_close=False):
+                            kde_bw=0.15, pca_component=0.90, plot_rmsd=True, save_dir=None,
+                            n_clusters="auto", eps=None, min_samples=None, metric="euclidean", 
+                            fig_close=False, num_cpus=None):
         """Analyze bound poses for binding sites.
 
         This function can find representative bound poses, cluster the bound poses and calculate pose RMSD for
@@ -808,13 +803,32 @@ class LipidInteraction:
         pose_format : str, default="pdb"
         score_weights : dict or None, default=None
         kde_bw : scalar, default=0.15
-        pca_component : int, float or ‘mle’, default=0.95
+        pca_component : int, float or ‘mle’, default=0.90
             Set the `n_components` value in `sklearn.decomposition.PCA`
         plot_rmsd : bool, default=True
         n_clusters : int or 'auto'
             If `n_clusters == 'auto'`, the number of clusters will be guessed based on density, whereas n_clusters=N
             will use KMeans to cluster the poses.
+        eps : float or None, default=None
+            The minimum neighbour distance used by `DBSCAN <https://scikit-learn.org/stable/modules/generated/sklearn.cluster.DBSCAN.html>`_
+            if None, the value will determined from as the elbow point of the sorted minimum neighbour distance 
+            of all the data points. 
+        min_samples : int or None, default=None
+            The minimum number of samples to be considered as core samples used by 
+            `DBSCAN <https://scikit-learn.org/stable/modules/generated/sklearn.cluster.DBSCAN.html>`_ .
+            If None, the value will be automatically determined based on the size of data. 
+        metric : str, default='euclidean'
+            The metric used to calculated neighbour distance used by 
+            `DBSCAN <https://scikit-learn.org/stable/modules/generated/sklearn.cluster.DBSCAN.html>`_ .
+            Default is the Euclidean distance. 
+        fig_close : bool, default=False
+            This parameter control whether to close the plotted figures using plt.close(). It can save memory if 
+            many figures are generated. 
+        num_cpus : int or None default=None
+            The number of CPUs used to rank the poses and cluster poses. Python multiprocessing deployed by 
+            `p_tqdm <https://github.com/swansonk14/p_tqdm>`_ is used to speed up these calculations. 
         save_dir : str or None, default=None
+        
 
         Returns
         -------
@@ -842,59 +856,32 @@ class LipidInteraction:
         pose_traj, pose_info = collect_bound_poses(selected_bs_map, self._contact_residues_low, self._trajfile_list,
                                                    self._topfile_list, self._lipid, self._protein_ref, self._lipid_ref,
                                                    stride=self._stride, nprot=self._nprot)
-        # analyize bound poses
-        RMSD_set = {}
-        for bs_id in tqdm(selected_bs_id, desc="ANALYZE BOUND POSES"):
-            # lipid_dist_per_atom shape: [n_lipid_atoms, n_bound_poses, n_BS_residues]
-            if len(pose_traj[bs_id]) == 0:
-                print(f"No Bound pose collected from Binding Site {bs_id}! Possibly due to insufficient sampling.")
-                continue
-            protein_atom_indices = [[atom.index for atom in residue.atoms]
-                                    for residue in self._protein_ref.top.residues]
-            lipid_atom_indices = [self._protein_ref.n_atoms + atom_idx
-                                  for atom_idx in np.arange(self._lipid_ref.n_atoms)]
-            lipid_dist_per_atom = vectorize_poses(pose_traj[bs_id], self._node_list[bs_id],
-                                                  protein_atom_indices, lipid_atom_indices)
-            if n_top_poses > 0:
-                pose_dir_rank = check_dir(pose_dir, "BSid{}_rank".format(bs_id), print_info=False)
-                atom_weights = {atom_idx: 1 for atom_idx in np.arange(self._lipid_ref.n_atoms)}
-                if score_weights is not None:
-                    translate = {atom_idx: score_weights[self._lipid_ref.atom(atom_idx)]
-                                 for atom_idx in np.arange(self._lipid_ref.n_atoms)
-                                 if self._lipid_ref.atom(atom_idx) in score_weights.keys()}
-                    atom_weights.update(translate)
-                scores = calculate_scores(lipid_dist_per_atom, kde_bw=kde_bw, pca_component=pca_component,
-                                          score_weights=atom_weights)
-                num_of_poses = min(n_top_poses, pose_traj[bs_id].n_frames)
-                pose_indices = np.argsort(scores)[::-1][:num_of_poses]
-                write_bound_poses(pose_traj[bs_id], pose_indices, pose_dir_rank, pose_prefix="BSid{}_top".format(bs_id),
-                                  pose_format=pose_format)
-                self._write_pose_info([pose_info[bs_id][pose_idx] for pose_idx in [pose_indices]],
-                                      f"{pose_dir_rank}/pose_info.txt")
-            lipid_dist_per_pose = np.array([lipid_dist_per_atom[:, pose_id, :].ravel()
-                                            for pose_id in np.arange(lipid_dist_per_atom.shape[1])])
-            # cluster poses
+        protein_atom_indices = [[atom.index for atom in residue.atoms]
+                                for residue in self._protein_ref.top.residues]
+        lipid_atom_indices = [self._protein_ref.n_atoms + atom_idx
+                              for atom_idx in np.arange(self._lipid_ref.n_atoms)]
+        atom_weights = {atom_idx: 1 for atom_idx in np.arange(self._lipid_ref.n_atoms)}
+        if score_weights is not None:
+            translate = {atom_idx: score_weights[self._lipid_ref.atom(atom_idx)]
+                         for atom_idx in np.arange(self._lipid_ref.n_atoms)
+                         if self._lipid_ref.atom(atom_idx) in score_weights.keys()}
+            atom_weights.update(translate)
 
-            pose_dir_clustered = check_dir(pose_dir, "BSid{}_clusters".format(bs_id), print_info=False)
-            transformed_data = PCA(n_components=pca_component).fit_transform(lipid_dist_per_pose)
-            if n_clusters == 'auto':
-                _, core_sample_indices = cluster_DBSCAN(transformed_data, eps=eps, min_samples=min_samples,
-                                                        metric=metric)
-                selected_pose_id = [np.random.choice(i_core_sample, 1)[0] for i_core_sample in core_sample_indices]
-            elif n_clusters > 0:
-                cluster_labels = cluster_KMeans(transformed_data, n_clusters=n_clusters)
-                cluster_id_set = np.unique(cluster_labels)
-                selected_pose_id = [np.random.choice(np.where(cluster_labels == cluster_id)[0], 1)[0]
-                                    for cluster_id in cluster_id_set]
-            write_bound_poses(pose_traj[bs_id], selected_pose_id, pose_dir_clustered,
-                              pose_prefix="BSid{}_cluster".format(bs_id), pose_format=pose_format)
-            self._write_pose_info([pose_info[bs_id][pose_idx] for pose_idx in [selected_pose_id]],
-                                  f"{pose_dir_clustered}/pose_info.txt")
-            # calculate pose rmsd
-            dist_mean = np.mean(lipid_dist_per_pose, axis=0)
-            RMSD_set["Binding Site {}".format(bs_id)] = [rmsd(lipid_dist_per_pose[pose_id], dist_mean)
-                                                         for pose_id in np.arange(len(lipid_dist_per_pose))]
-            pose_rmsd_per_residue[self._node_list[bs_id]] = np.mean(RMSD_set["Binding Site {}".format(bs_id)])
+        if n_top_poses > 0:
+            # multiprocessing wrapped under p_tqdm
+            rmsd_set = p_map(partial(analyze_pose_wrapper, protein_atom_indices=protein_atom_indices,
+                                     lipid_atom_indices=lipid_atom_indices, n_top_poses=n_top_poses,
+                                     pose_dir=pose_dir, atom_weights=atom_weights, kde_bw=kde_bw,
+                                     pca_component=pca_component, pose_format=pose_format, n_clusters=n_clusters,
+                                     eps=eps, min_samples=min_samples, metric=metric,
+                                     trajfile_list=self._trajfile_list),
+                             selected_bs_id, [pose_traj[bs_id] for bs_id in selected_bs_id],
+                             [self._node_list[bs_id] for bs_id in selected_bs_id],
+                             [pose_info[bs_id] for bs_id in selected_bs_id], num_cpus=num_cpus, desc="ANALYZE BOUND POSES")
+            RMSD_set = {}
+            for bs_id, rmsd in zip(selected_bs_id, rmsd_set):
+                RMSD_set["Binding Site {}".format(bs_id)] = rmsd
+                pose_rmsd_per_residue[self._node_list[bs_id]] = np.mean(RMSD_set["Binding Site {}".format(bs_id)])
         # update dataset
         self.dataset["Binding Site Pose RMSD"] = pose_rmsd_per_residue
         pose_rmsd_data = pd.DataFrame(
@@ -905,7 +892,8 @@ class LipidInteraction:
                                    title="{}".format(self._lipid), ylabel="RMSD (nm)", fig_close=fig_close)
         return pose_traj, pose_rmsd_data
 
-    def compute_surface_area(self, binding_site_id=None, radii=None, plot_data=True, save_dir=None, fig_close=False):
+    def compute_surface_area(self, binding_site_id=None, radii=None, plot_data=True, save_dir=None,
+                             fig_close=False, num_cpus=None):
         """Calculate binding site surface areas.
 
         Parameters
@@ -914,6 +902,12 @@ class LipidInteraction:
         radii : dict or None, default=None
         plot_data : bool, default=True
         save_dir : str or None, default=None
+        fig_close : bool, default=False
+        num_cpus : int or None, default=None
+
+        Returns
+        -----------
+        surface_area : pandas.DataFrame
 
         """
         MARTINI_CG_radii = {"BB": 0.26, "SC1": 0.23, "SC2": 0.23, "SC3": 0.23}
@@ -934,9 +928,18 @@ class LipidInteraction:
         selected_bs_id = np.atleast_1d(np.array(binding_site_id, dtype=int)) if binding_site_id is not None \
             else np.arange(len(self._node_list), dtype=int)
         selected_bs_id_map = {bs_id: self._node_list[bs_id] for bs_id in selected_bs_id}
-        surface_area_data = calculate_site_surface_area(selected_bs_id_map, radii_book, self._trajfile_list,
-                                                        self._topfile_list, self._nprot, self._timeunit, self._stride,
-                                                        dt_traj=self._dt_traj)
+        returned_values = p_map(partial(calculate_surface_area_wrapper, binding_site_map=selected_bs_id_map,
+                                        nprot=self._nprot, timeunit=self._timeunit, stride=self._stride,
+                                        dt_traj=self._dt_traj, radii_book=radii_book), self._trajfile_list,
+                                self._topfile_list, np.arange(len(self._trajfile_list), dtype=int),
+                                num_cpus=num_cpus, desc="CALCULATE BINDING SITE SURFACE AREA")
+        surface_data = []
+        data_keys = []
+        for returned_tuple in returned_values:
+             for idx in np.arange(len(returned_tuple[0])):
+                surface_data.append(returned_tuple[0][idx])
+                data_keys.append(returned_tuple[1][idx])
+        surface_area_data = pd.concat(surface_data, keys=data_keys)
         # update dataset
         for bs_id in selected_bs_id:
             nodes = self._node_list[bs_id]
@@ -1137,7 +1140,12 @@ class LipidInteraction:
             if fn is None:
                 fn = "BindingSites_Info_{}.txt".format(self._lipid)
             self._check_calculation("Binding Site ID", self.compute_binding_nodes, print_data=False)
-            self._check_calculation("Binding Site Koff", self.compute_site_koff, print_data=False)
+            self._check_calculation("Binding Site Koff", self.compute_site_koff, plot_data=False)
+            mapping_funcs = {"Residence Time": self.compute_residue_koff,
+                             "Duration": self.compute_residue_duration,
+                             "Occupancy": self.compute_residue_occupancy,
+                             "Lipid Count": self.compute_residue_lipidcount}
+            self._check_calculation(sort_residue, mapping_funcs[sort_residue])
             with open(os.path.join(BS_dir, fn), "a") as f:
                 f.write(f"## Network modularity {self._network_modularity:5.3f}")
                 f.write("\n")
@@ -1218,22 +1226,6 @@ class LipidInteraction:
     #     assisting func
     ############################################
 
-    def _format_koff_text(self, properties):
-        """Format text for koff plot. """
-        tu = "ns" if self._timeunit == "ns" else r"$\mu$s"
-        text = "{:18s} = {:.3f} {:2s}$^{{-1}} $\n".format("$k_{{off1}}$", properties["ks"][0], tu)
-        text += "{:18s} = {:.3f} {:2s}$^{{-1}} $\n".format("$k_{{off2}}$", properties["ks"][1], tu)
-        text += "{:14s} = {:.4f}\n".format("$R^2$", properties["r_squared"])
-        ks_boot_avg = np.mean(properties["ks_boot_set"], axis=0)
-        cv_avg = 100 * np.std(properties["ks_boot_set"], axis=0) / np.mean(properties["ks_boot_set"], axis=0)
-        text += "{:18s} = {:.3f} {:2s}$^{{-1}}$ ({:3.1f}%)\n".format("$k_{{off1, boot}}$", ks_boot_avg[0],
-                                                                     tu, cv_avg[0])
-        text += "{:18s} = {:.3f} {:2s}$^{{-1}}$ ({:3.1f}%)\n".format("$k_{{off2, boot}}$", ks_boot_avg[1],
-                                                                     tu, cv_avg[1])
-        text += "{:14s} = {:.4f}\n".format("$R^2$$_{{boot}}$", np.mean(properties["r_squared_boot_set"]))
-        text += "{:18s} = {:.3f} {:2s}".format("$Res. Time$", properties["res_time"], tu)
-        return text
-
     def _format_BS_print_info(self, bs_id, nodes, sort_item):
         """Format binding site information."""
         Residue_property_book = {"ARG": "Pos. Charge", "HIS": "Pos. Charge", "LYS": "Pos. Charge",
@@ -1295,14 +1287,3 @@ class LipidInteraction:
             print("Finished the calculation of {}.".format(item))
             print("#" * 60)
 
-    def _write_pose_info(self, selected_pose_info, fn):
-        """Write pose information for each selected pose. """
-        with open(fn, "w") as f:
-            for idx, info in enumerate(selected_pose_info):
-                f.write("POSE ID : {}\n".format(int(idx)))
-                f.write("TRAJ FN : {}\n".format(self._trajfile_list[info[0]]))
-                f.write("PROT ID : {}\n".format(info[1]))
-                f.write("LIPID ID: {}\n".format(info[2]))
-                f.write("TIME    : {:8.3f} ps\n".format(info[3]))
-                f.write("\n")
-                f.write("\n")
\ No newline at end of file
